/*
* DHT11/DHT22 bit banging GPIO driver
*
* Copyright (c) Harald Geyer <harald@ccbib.org>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*/

#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/printk.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/sysfs.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/wait.h>
#include <linux/bitops.h>
#include <linux/completion.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include <linux/of_gpio.h>

#include <linux/iio/iio.h>

#include "dht11_internal.h"

#define DRIVER_NAME	"dht11"

#define DHT11_DATA_VALID_TIME	2000000000  /* 2s in ns */

#define DHT11_EDGES_PREAMBLE 2
#define DHT11_BITS_PER_READ 40
#define DHT11_EDGES_PER_READ (2*DHT11_BITS_PER_READ + DHT11_EDGES_PREAMBLE + 1)

/* Data transmission timing (nano seconds) */
#define DHT11_START_TRANSMISSION_MIN 8000 /* us */
#define DHT11_START_TRANSMISSION_MAX 9000 /* us */
#define DHT11_WAIT_TRANSMISSION_MIN	10000  /* us */
#define DHT11_WAIT_TRANSMISSION_MAX	11000  /* us */

#define DHT11_MIN_TIMERES	34000  /* ns */
#define DHT11_THRESHOLD		49000  /* ns */
#define DHT11_AMBIG_LOW		23000  /* ns */
#define DHT11_AMBIG_HIGH	30000  /* ns */

struct dht11 {
	struct device			*dev;

	int				gpio;
	int				irq;

	struct completion		completion;

	s64				timestamp;
	int				temperature;
	int				humidity;

	/* num_edges: -1 means "no transmission in progress" */
	int				num_edges;
	struct {s64 ts; int value; }	edges[DHT11_EDGES_PER_READ];
};

/**
* reinit_completion - reinitialize a completion structure
* @x:  pointer to completion structure that is to be reinitialized
*
* This inline function should be used to reinitialize a completion structure so it can
* be reused. This is especially important after complete_all() is used.
*/
static inline void reinit_completion(struct completion *x)
{
	x->done = 0;
}

static void devm_iio_device_release(struct device *dev, void *res)
{
	iio_device_free(*(struct iio_dev **)res);
}

static int devm_iio_device_match(struct device *dev, void *res, void *data)
{
	struct iio_dev **r = res;
	if (!r || !*r) {
		WARN_ON(!r || !*r);
		return 0;
	}
	return *r == data;
}

/**
* devm_iio_device_alloc - Resource-managed iio_device_alloc()
* @dev:		Device to allocate iio_dev for
* @sizeof_priv:	Space to allocate for private structure.
*
* Managed iio_device_alloc. iio_dev allocated with this function is
* automatically freed on driver detach.
*
* If an iio_dev allocated with this function needs to be freed separately,
* devm_iio_device_free() must be used.
*
* RETURNS:
* Pointer to allocated iio_dev on success, NULL on failure.
*/
struct iio_dev *devm_iio_device_alloc(struct device *dev, int sizeof_priv)
{
	struct iio_dev **ptr, *iio_dev;

	ptr = devres_alloc(devm_iio_device_release, sizeof(*ptr),
		GFP_KERNEL);
	if (!ptr)
		return NULL;

	/* use raw alloc_dr for kmalloc caller tracing */
	iio_dev = iio_device_alloc(sizeof_priv);
	if (iio_dev) {
		*ptr = iio_dev;
		devres_add(dev, ptr);
	} else {
		devres_free(ptr);
	}

	return iio_dev;
}

/**
* devm_iio_device_free - Resource-managed iio_device_free()
* @dev:		Device this iio_dev belongs to
* @iio_dev:		the iio_dev associated with the device
*
* Free iio_dev allocated with devm_iio_device_alloc().
*/
void devm_iio_device_free(struct device *dev, struct iio_dev *iio_dev)
{
	int rc;

	rc = devres_release(dev, devm_iio_device_release,
		devm_iio_device_match, iio_dev);
	WARN_ON(rc);
}


static void devm_iio_device_unreg(struct device *dev, void *res)
{
	iio_device_unregister(*(struct iio_dev **)res);
}

/**
* devm_iio_device_register - Resource-managed iio_device_register()
* @dev:	Device to allocate iio_dev for
* @indio_dev:	Device structure filled by the device driver
*
* Managed iio_device_register.  The IIO device registered with this
* function is automatically unregistered on driver detach. This function
* calls iio_device_register() internally. Refer to that function for more
* information.
*
* If an iio_dev registered with this function needs to be unregistered
* separately, devm_iio_device_unregister() must be used.
*
* RETURNS:
* 0 on success, negative error number on failure.
*/
int devm_iio_device_register(struct device *dev, struct iio_dev *indio_dev)
{
	struct iio_dev **ptr;
	int ret;

	ptr = devres_alloc(devm_iio_device_unreg, sizeof(*ptr), GFP_KERNEL);
	if (!ptr)
		return -ENOMEM;

	*ptr = indio_dev;
	ret = iio_device_register(indio_dev);
	if (!ret)
		devres_add(dev, ptr);
	else
		devres_free(ptr);

	return ret;
}

/**
* devm_iio_device_unregister - Resource-managed iio_device_unregister()
* @dev:	Device this iio_dev belongs to
* @indio_dev:	the iio_dev associated with the device
*
* Unregister iio_dev registered with devm_iio_device_register().
*/
void devm_iio_device_unregister(struct device *dev, struct iio_dev *indio_dev)
{
	int rc;

	rc = devres_release(dev, devm_iio_device_unreg,
		devm_iio_device_match, indio_dev);
	WARN_ON(rc);
}

static int gpiopin = -1;
module_param_named(gpio, gpiopin, int, 0444); 
MODULE_PARM_DESC(gpio, "DHT GPIO pin number"); 
module_param_named(debug, gpiodebug, int, 0444); 
MODULE_PARM_DESC(debug, "DHT module debug level (0 - 3)"); 

static unsigned char dht11_decode_byte(char *bits)
{
	unsigned char ret = 0;
	int i;

	for (i = 0; i < 8; ++i) {
		ret <<= 1;
		if (bits[i])
			++ret;
	}

	return ret;
}

static int dht11_decode(struct dht11 *dht11, int offset)
{
	int i, t;
	char bits[DHT11_BITS_PER_READ];
	unsigned char temp_int, temp_dec, hum_int, hum_dec, checksum;

	for (i = 0; i < DHT11_BITS_PER_READ; ++i) {
		t = dht11->edges[offset + 2 * i + 2].ts -
			dht11->edges[offset + 2 * i + 1].ts;
		if (!dht11->edges[offset + 2 * i + 1].value) {
			if(gpiodebug){
				printk("dht11: lost synchronisation at edge %d\n",
					offset + 2 * i + 1);
			}
			return -EIO;
		}
		bits[i] = t > DHT11_THRESHOLD;
	}

	hum_int = dht11_decode_byte(bits);
	hum_dec = dht11_decode_byte(&bits[8]);
	temp_int = dht11_decode_byte(&bits[16]);
	temp_dec = dht11_decode_byte(&bits[24]);
	checksum = dht11_decode_byte(&bits[32]);

	if (((hum_int + hum_dec + temp_int + temp_dec) & 0xff) != checksum) {
		if(gpiodebug){
			printk("dht11: invalid checksum\n");
		}
		return -EIO;
	}

	dht11->timestamp = iio_get_time_ns();
	if (hum_int < 20) {  /* DHT22 */
		dht11->temperature = (((temp_int & 0x7f) << 8) + temp_dec) *
			((temp_int & 0x80) ? -100 : 100);
		dht11->humidity = ((hum_int << 8) + hum_dec) * 100;
	} else if (temp_dec == 0 && hum_dec == 0) {  /* DHT11 */
		dht11->temperature = temp_int * 1000;
		dht11->humidity = hum_int * 1000;
	} else {
		if(gpiodebug){
			printk("Don't know how to decode data: %d %d %d %d\n",
				hum_int, hum_dec, temp_int, temp_dec);
		}
		return -EIO;
	}

	return 0;
}

/*
* IRQ handler called on GPIO edges
*/
static irqreturn_t dht11_handle_irq(int irq, void *data)
{
	struct iio_dev *iio = data;
	struct dht11 *dht11 = iio_priv(iio);
	int state; 

	/* TODO: Consider making the handler safe for IRQ sharing */
	if (dht11->num_edges < DHT11_EDGES_PER_READ && dht11->num_edges >= 0) {
		dht11->edges[dht11->num_edges].ts = iio_get_time_ns();

		state = gpio_get_value(dht11->gpio);
		dht11->edges[dht11->num_edges++].value = state;
		if (dht11->num_edges >= DHT11_EDGES_PER_READ)
			complete(&dht11->completion);
	}

	return IRQ_HANDLED;
}

static int dht11_read_raw(struct iio_dev *iio_dev,
	const struct iio_chan_spec *chan,
	int *val, int *val2, long m)
{
	struct dht11 *dht11 = iio_priv(iio_dev);
	int ret, offset;

	if (dht11->timestamp + DHT11_DATA_VALID_TIME < iio_get_time_ns()) {
		reinit_completion(&dht11->completion);

		dht11->num_edges = 0;
		ret = gpio_direction_output(dht11->gpio, 0); // read signal, at least 1ms
		if (ret)
			goto err;
		usleep_range(DHT11_START_TRANSMISSION_MIN, DHT11_START_TRANSMISSION_MAX); 

		ret = gpio_direction_input(dht11->gpio);
		if (ret)
			goto err;

		ret = gpio_set_pull(dht11->gpio, 1); // set PullUp
		if (ret)
			goto err;

		ret = request_irq(dht11->irq, dht11_handle_irq,
			IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			iio_dev->name, iio_dev);
		if (ret){
			dev_err(&iio_dev->dev, "Failed to allocate IRQ. err=%d.\n", ret);
			goto err;
		}

		ret = gpio_set_pin_mode(dht11->gpio, 6); // enable EINT mode
		if (ret)
			goto err;

		ret = gpio_set_isr(dht11->gpio, 1);
		if (ret)
			goto err;

		gpio_set_hsoc(1);

		// it is recommended to use usleep_range for delays < 20 ms
		usleep_range(DHT11_WAIT_TRANSMISSION_MIN, DHT11_WAIT_TRANSMISSION_MAX); 

		ret = wait_for_completion_killable_timeout(&dht11->completion, HZ);

		free_irq(dht11->irq, iio_dev);

		if (ret == 0 && dht11->num_edges < DHT11_EDGES_PER_READ - 1) {
			dev_err(&iio_dev->dev,
				"Only %d signal edges detected\n",
				dht11->num_edges);
			ret = -ETIMEDOUT;
		}
		if (ret < 0)
			goto err;

		offset = DHT11_EDGES_PREAMBLE +
			dht11->num_edges - DHT11_EDGES_PER_READ;
		for (; offset >= 0; --offset) {
			ret = dht11_decode(dht11, offset);
			if (!ret)
				break;
		}

		if (ret)
			goto err;
	}

	ret = IIO_VAL_INT;
	if (chan->type == IIO_TEMP)
		*val = dht11->temperature;
	else if (chan->type == IIO_HUMIDITYRELATIVE)
		*val = dht11->humidity;
	else
		ret = -EINVAL;
err:
	gpio_set_hsoc(0);
	dht11->num_edges = -1;
	return ret;
}

static const struct iio_info dht11_iio_info = {
	.driver_module		= THIS_MODULE,
	.read_raw		= dht11_read_raw,
};


static const struct iio_chan_spec dht11_chan_spec[] = {
	{ .type = IIO_TEMP,
	.info_mask_separate = BIT(IIO_CHAN_INFO_PROCESSED), },
	{ .type = IIO_HUMIDITYRELATIVE,
	.info_mask_separate = BIT(IIO_CHAN_INFO_PROCESSED), }
};

static const struct of_device_id dht11_dt_ids[] = {
	{ .compatible = "dht11", },
	{ }
};
/* MODULE_DEVICE_TABLE(of, dht11_dt_ids); */


/*
Note: Not all GPIO pins have interrupts. This is hardware limitation. 
Allwinner R80: only CPU pins from PB, PG, PH and PL banks have interrupts. 
CPU pin numbers: 32...42, 192...205, 224...235, 352...365
*/
static int dht11_probe(struct platform_device *pdev)
{
	struct dht11 *dht11;
	struct iio_dev *iio;
	int ret;
	struct device *dev = &pdev->dev;
	struct device_node *node = dev->of_node;

	if(gpiodebug > 1){
		printk("dht11: probe(). Default pin number: %i\n", gpiopin);
	}

	iio = devm_iio_device_alloc(dev, sizeof(*dht11));
	if (!iio) {
		dev_err(dev, "Failed to allocate IIO device\n");
		return -ENOMEM;
	}
	if(gpiodebug){
		printk("dht11: Device allocated\n");
	}

	dht11 = iio_priv(iio);
	dht11->dev = dev;

	if(gpiopin > 0)
	{
		printk("dht11: Using pin number from the command line: %i\n", gpiopin);
		dht11->gpio = gpiopin;
	}
	else
	{
		dht11->gpio = ret = of_get_gpio(node, 0);
		if (ret < 0) {
			dev_err(dev, "Failed to get a pin number. Try 'gpio=<num> parameter\n");
			return ret;
		}
	}
	ret = devm_gpio_request_one(dev, dht11->gpio, GPIOF_IN, pdev->name);
	if (ret) {
		dev_err(dev, "Failed to allocate GPIO pin %d. err=%d.\n",
			dht11->gpio, ret);
		return ret;
	}

	dht11->irq = gpio_to_irq(dht11->gpio);
	if (dht11->irq < 0) {
		dev_err(dev, "GPIO %d has no interrupt\n", dht11->gpio);
		return -EINVAL;
	}

	if(gpiodebug > 1){
		printk("dht11: GPIO pin IRQ %d\n", dht11->irq);
	}

	dht11->timestamp = iio_get_time_ns() - DHT11_DATA_VALID_TIME - 1;
	dht11->num_edges = -1;

	platform_set_drvdata(pdev, iio);

	init_completion(&dht11->completion);
	iio->name = pdev->name;
	iio->dev.parent = &pdev->dev;
	iio->info = &dht11_iio_info;
	iio->modes = INDIO_DIRECT_MODE;
	iio->channels = dht11_chan_spec;
	iio->num_channels = ARRAY_SIZE(dht11_chan_spec);

	if(gpiodebug > 1){
		printk("dht11: devm_iio_device_register()\n");
	}

	return devm_iio_device_register(dev, iio);
}

/* Function to deinitialize the processing for usage id */
static int dht11_remove(struct platform_device *pdev)
{
	if(gpiodebug > 1){
		printk("dht11: remove()\n");
	}
	return 0;
}

static void dht11_release(struct device *dev)
{
	if(gpiodebug > 1){
		printk("dht11: release()\n");
	}
}

static struct platform_driver dht11_driver = {
	.driver = {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = dht11_dt_ids,
	},
	.probe  = dht11_probe,
	.remove  = dht11_remove,
};

static struct platform_device dht11_device = {
	.name = DRIVER_NAME,
	.id   = -1,
	.dev    = {
		.release        = dht11_release,
	}
}; 

/* ---------------------------------------------------------------------
* Module setup/teardown
*/

static int __init dht11_init(void)
{
	int rc;

	if(gpiodebug > 1){
		printk("dht11: init()\n");
	}

	rc = platform_driver_register(&dht11_driver);
	if (!rc) {
		if(gpiodebug > 1){
			printk("dht11: platform_device_register()\n");
		}
		rc = platform_device_register(&dht11_device);
		if (rc)
			platform_driver_unregister(&dht11_driver);
	}
	return rc;
}

static void __exit dht11_exit(void)
{
	if(gpiodebug > 1){
		printk("dht11: exit()\n");
	}

	platform_device_unregister(&dht11_device);
	platform_driver_unregister(&dht11_driver);
}  

module_init(dht11_init);
module_exit(dht11_exit); 

MODULE_AUTHOR("Harald Geyer <harald@ccbib.org>");
MODULE_DESCRIPTION("DHT11 humidity/temperature sensor driver");
MODULE_LICENSE("GPL v2");